/*
 * Entropy.h
 *
 *  Created on: Jul 6, 2009
 *      Author: Alessio Sclocco <alessio@sclocco.eu>
 *
 *     License: See the end of this file for license information
 */

#ifndef ENTROPY_H_
#define ENTROPY_H_

#include "AleatoryVariable.h"
#include "SampleSet.h"

#include <cmath>
#include <set>
#include <map>
#include <list>

using std::set;
using std::map;
using std::pair;
using std::list;

namespace ita {

template< class T > double entropy(AleatoryVariable< T > &var);
template< class T > double jointEntropy(set< unsigned int > &x, SampleSet &sset);
template< class T > double conditionalEntropy(set< unsigned int > &x, set< unsigned int > &y, SampleSet &sset);
template< class T > double mutualInformation(set< unsigned int > &x, set< unsigned int > &y, SampleSet &sset);

template< class T > double entropy(AleatoryVariable< T > &var) {
	T symbol;
	double tempEntropy = 0;

	var.resetNextSymbol();
	try {
		while ( true ) {
			symbol = var.getNextSymbol();
			tempEntropy += var.getSymbolProbability(symbol) * log2(var.getSymbolProbability(symbol));
		}
	} catch ( ... ) {
		var.resetNextSymbol();
	}

	return (tempEntropy != 0 ) ? -tempEntropy : tempEntropy;
}

template< class T > double jointEntropy(set< unsigned int > &x, SampleSet &sset) {
	map< vector< T >, unsigned int > symbolCounter;
	vector< AleatoryVariable< T > * > vars;
	list< vector< T > > sequence;
	unsigned int samples = 0;

	for ( set< unsigned int >::const_iterator setIterator = x.begin(); setIterator != x.end(); setIterator++ ) {
		vars.push_back(sset.getVariableById(*setIterator));
	}

	for ( unsigned int i(0); i < (vars[0])->getSequenceLength(); i++ ) {
		vector< T > tempSymbol;
		for ( typename vector< AleatoryVariable< T > * >::const_iterator varIterator = vars.begin(); varIterator != vars.end(); varIterator++ ) {
			tempSymbol.push_back((*varIterator)->getInSequence(i));
		}
		sequence.push_back(tempSymbol);
	}

	bool twoOnHead = false;
	for ( typename list< vector< T > >::iterator listIter = sequence.begin(); listIter != sequence.end(); listIter++ ) {
		if ( twoOnHead ) {
			listIter--;
			twoOnHead = false;
		}

		bool good = true;
		for ( unsigned int i(0); i < (*listIter).size(); i++ ) {
			if ( (*listIter)[i] == sset.getNullSymbol() ) {
				(vars[i])->resetNextSymbol();
				try {
					while ( true ) {
						vector< T > newVec = (*listIter);
						newVec[i] = (vars[i])->getNextSymbol();
						sequence.push_back(newVec);
					}
				} catch ( ... ) {
					(vars[i])->resetNextSymbol();
				}
				good = false;
				break;
			}
		}

		if ( good ) {
			typename map< vector< T >, unsigned int >::iterator symbolPosition;
			if ( (symbolPosition = symbolCounter.find(*listIter)) == symbolCounter.end() ) {
				symbolCounter.insert(pair< vector< T >, unsigned int >(*listIter, 1));
			}
			else {
				(*symbolPosition).second++;
			}
			samples++;
		}

		listIter = sequence.erase(listIter);
		if ( listIter != sequence.begin() ) {
			listIter--;
		}
		else {
			twoOnHead = true;
		}
	}

	AleatoryVariable< vector< T > > newVar;
	for( typename map< vector< T >, unsigned int >::const_iterator symbolIterator = symbolCounter.begin(); symbolIterator != symbolCounter.end(); symbolIterator++ ) {
		newVar.addSymbol((*symbolIterator).first, static_cast< double >((*symbolIterator).second) / samples);
	}

	return entropy< vector< T > >(newVar);
}

template< class T > double conditionalEntropy(set< unsigned int > &x, set< unsigned int > &y, SampleSet &sset) {
	set< unsigned int >	both;

	if ( x.size() > y.size() ) {
		both = x;
		for ( set< unsigned int >::const_iterator yIter = y.begin(); yIter != y.end(); yIter++ ) {
			both.insert(*yIter);
		}
	}
	else {
		both = y;
		for ( set< unsigned int >::const_iterator xIter = x.begin(); xIter != x.end(); xIter++ ) {
			both.insert(*xIter);
		}
	}

	return jointEntropy< T >(both, sset) - jointEntropy< T >(y, sset);
}

template< class T > double mutualInformation(set< unsigned int > &x, set< unsigned int > &y, SampleSet &sset) {
	return jointEntropy< T >(x, sset) - conditionalEntropy< T >(x, y, sset);
}

}

#endif /* ENTROPY_H_ */

/*
 * Copyright 2009 Alessio Sclocco <alessio@sclocco.eu>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
